@@ -1,59 +1,11 @@
 """All types used by funml"""
 from inspect import signature
-from typing import Any, Type, Union, Callable, Optional, List, Tuple
+from typing import Any, Union, Callable, Optional, List, Tuple
 
 from funml import errors
 from funml.utils import is_equal_or_of_type
 
 
-class Assignment:
-    """A variable assignment
-
-    Assigns a given value a variable name and type. It will check that
-    the data type passed is as expected. It can thus be used to
-    validate third party data before passing it through the ml-program.
-
-    Args:
-        var: the variable name
-        t: the variable type
-        val: the value stored in the variable
-
-    Raises:
-        TypeError: `val` passed is not of type `t`
-    """
-
-    def __init__(self, var: Any, t: Type = type(None), val: Any = None):
-        self.__var = var
-        self.__t = t
-
-        if not isinstance(val, t):
-            raise TypeError(f"expected type {t}, got {type(val)}")
-
-        self.__val = val
-
-    def __rshift__(self, nxt: Union["Expression", "Assignment", Callable]):
-        """This makes piping using the '>>' symbol possible
-
-        Combines with the given expression, assignments, Callables to produce a new expression
-        where data flows from current to nxt
-        """
-        return _append_expn(self, nxt)
-
-    def __iter__(self):
-        """Generates an iterator that can be used to create a dict using dict()"""
-        yield self.__var, self.__val
-
-    def __call__(self) -> Any:
-        """Returns the value associated with this assignment"""
-        return self.__val
-
-
-class Context(dict):
-    """The context map containing variables in scope."""
-
-    pass
-
-
 class MLType:
     """An ML-enabled type, that can easily be used in pattern matching, piping etc.
 
@@ -84,6 +36,79 @@ class MLType:
         raise NotImplemented("_is_like not implemented")
 
 
+class Pipeline:
+    """A series of logic blocks that operate on the same data in sequence.
+
+    This has internal state so it is not be used in such stuff as recursion.
+    However when compile is run on it, a reusable (pure) expression is created.
+    """
+
+    def __init__(self):
+        self._queue: List[Expression] = []
+        self._is_terminated = False
+
+    def __rshift__(self, nxt: Union["Expression", Callable, "Pipeline"]):
+        """Uses `>>` to append the nxt expression, callable, pipeline to this pipeline.
+
+        Args:
+            nxt: the next expression, pipeline, or callable to apply after the current one.
+
+        Raises:
+            ValueError: when the pipeline is already terminated with ml.execute() in its queue.
+        """
+        self.__update_queue(nxt)
+        if self._is_terminated:
+            return self()
+
+        return self
+
+    def __call__(self, *args: Any, **kwargs: Any) -> Any:
+        """Computes the logic within the pipeline and returns the value.
+
+        This method runs all those expressions in the queue sequentially,
+        with the output of an expression being used as
+        input for the next expression.
+
+        Args:
+            args: any arguments passed.
+            kwargs: any key-word arguments passed
+
+        Returns:
+            the computed output of this pipeline.
+        """
+        output = None
+        queue = self._queue[:-1] if self._is_terminated else self._queue
+
+        for expn in queue:
+            if output is None:
+                output = expn(*args, **kwargs)
+            else:
+                # make sure piped expressions only consume previous outputs args, and kwargs
+                output = expn(output, **kwargs)
+
+        return output
+
+    def __copy__(self):
+        """Helps call copy on a pipeline"""
+        new_pipeline = Pipeline()
+        new_pipeline._queue += self._queue
+        new_pipeline._is_terminated = self._is_terminated
+        return new_pipeline
+
+    def __update_queue(self, nxt):
+        """Appends a pipeline or an expression to the queue."""
+        if self._is_terminated:
+            raise ValueError("a terminated pipeline cannot be extended.")
+
+        if isinstance(nxt, Pipeline):
+            self._queue += nxt._queue
+            self._is_terminated = nxt._is_terminated
+        else:
+            nxt_expn = to_expn(nxt)
+            self._queue.append(nxt_expn)
+            self._is_terminated = isinstance(nxt, ExecutionExpression)
+
+
 class Expression:
     """Logic that returns a value when applied.
 
@@ -94,16 +119,8 @@ class Expression:
         f: the operation or logic to run as part of this expression
     """
 
-    # FIXME: Expressions are not working right when undergoing recursion
-    #       It seems as if the operation f is frozen and never changes
-    #       or something like that. Try:     accum_factorial = ml.val(lambda num, accum: (
-    #         ml.match(num <= 0).case(True, do=lambda: accum).case(False, accum_factorial(num - 1, num * accum))()
-    #     ))
-    #     factorial = ml.val(lambda x: accum_factorial(x, 1))
     def __init__(self, f: Optional["Operation"] = None):
         self._f = f if f is not None else Operation(lambda x, *args, **kwargs: x)
-        self._context: "Context" = Context()
-        self._queue: List[Expression] = []
 
     def __call__(self, *args: Any, **kwargs: Any) -> Any:
         """Computes the logic within and returns the value.
@@ -115,72 +132,39 @@ class Expression:
         Returns:
             the computed output of this expression.
         """
-        prev_output = self._run_prev_expns(*args, **kwargs)
+        return self._f(*args, **kwargs)
 
-        if isinstance(prev_output, Context):
-            self._context.update(prev_output)
-        elif prev_output is not None:
-            # make sure piped expressions only consume previous outputs args, and kwargs
-            return self._f(prev_output, **self._context, **kwargs)
-
-        return self._f(*args, **self._context, **kwargs)
-
-    def __rshift__(self, nxt: Union["Expression", "Assignment", Callable]):
+    def __rshift__(self, nxt: Union["Expression", "Pipeline", Callable]) -> "Pipeline":
         """This makes piping using the '>>' symbol possible.
 
-        Combines with the given `nxt` expression to produce a new expression
+        Combines with the given `nxt` expression or pipeline to produce a new pipeline
         where data flows from current to nxt.
 
         Args:
-            nxt: the next expression, assignment or callable to apply after the current one.
+            nxt: the next expression, pipeline, or callable to apply after the current one.
+
+        Returns:
+            a new pipeline where the first expression is the current expression followed by `nxt`
         """
-        merged_expn = _append_expn(self, nxt)
+        new_pipeline = Pipeline()
+        new_pipeline >> self >> nxt
+        return new_pipeline
 
-        if isinstance(nxt, ExecutionExpression):
-            # stop pipeline, execute and return values
-            return merged_expn()
 
-        return merged_expn
+class ExecutionExpression(Expression):
+    """Expression that executes all previous once it is found on a pipeline
 
-    def _run_prev_expns(self, *args: Any, **kwargs: Any) -> Union["Context", Any]:
-        """Runs all the previous expressions, returning the final output.
+    Raises:
+        NotImplementedError: when `>>` is used after it.
+    """
 
-        In order to have expressions piped, all expressions are queued in the
-        expression at the end of the pipe. This method runs all those expressions sequentially,
-        with the output of the previous expression being used as input of the current expression.
+    def __rshift__(self, nxt: Any):
+        """rshift is not supported for this.
 
-        Args:
-            args: Any args passed
-            kwargs: Any key-word arguments passed.
-
-         Returns:
-             The output of the most previous expression, after running all expressions before it and piping their output
-             sequentially to the next.
+        This is a terminal expression that expects no other expression
+        after it on the pipeline.
         """
-        output = None
-
-        for expn in self._queue:
-            if output is None:
-                output = expn(*args, **expn._context, **kwargs)
-            elif isinstance(output, Context):
-                output = expn(*args, **expn._context, **output, **kwargs)
-            else:
-                # make sure piped expressions only consume previous outputs args, and kwargs
-                output = expn(output, **expn._context, **kwargs)
-
-        return output
-
-    def append_prev_expns(self, *expns: "Expression"):
-        """Appends expressions that should be computed before this one.
-
-        Args:
-            expns: the previous expressions to add to the queue
-        """
-        self._queue += expns
-
-    def clear_prev_expns(self):
-        """Clears all previous expressions in queue."""
-        self._queue.clear()
+        raise NotImplementedError("terminal pipeline expression: `>>` not supported")
 
 
 class MatchExpression(Expression):
@@ -191,7 +175,7 @@ class MatchExpression(Expression):
     """
 
     def __init__(self, arg: Optional[Any] = None):
-        super().__init__(f=Operation(self))
+        super().__init__()
         self._matches: List[Tuple[Callable, Expression]] = []
         self.__arg = arg
 
@@ -199,7 +183,7 @@ class MatchExpression(Expression):
         """Adds a case to a match statement.
 
         This is chainable, allowing multiple cases to be added to the same
-        match expression.
+        match pipeline.
 
         Args:
             pattern: the pattern to match against.
@@ -217,12 +201,12 @@ class MatchExpression(Expression):
         self.__add_match(check=check, expn=expn)
         return self
 
-    def __add_match(self, check: Callable, expn: Expression):
-        """Adds a match set to the list of matches.
+    def __add_match(self, check: Callable, expn: "Expression"):
+        """Adds a match set to the list of match sets
 
         A match set comprises a checker function and an expression.
         The checker function checks if a given argument matches this case.
-        The expression that is called when the case is matched.
+        The expression is called when the case is matched.
 
         Args:
             check: the checker function
@@ -241,7 +225,8 @@ class MatchExpression(Expression):
     def __call__(self, arg: Optional[Any] = None) -> Any:
         """Applies the matched case and returns the output.
 
-        The match cases are surveyed for any that matches the given argument until one that matches is found.
+        The match cases are surveyed for any that matches the given argument
+        until one that matches is found.
         Then the expression of that case is run and its output returned.
 
         Args:
@@ -256,11 +241,6 @@ class MatchExpression(Expression):
         if arg is None:
             arg = self.__arg
 
-        args = [] if arg is None else [arg]
-        prev_output = self._run_prev_expns(*args)
-        if prev_output is not None:
-            arg = prev_output
-
         for check, expn in self._matches:
             if check(arg):
                 return expn(arg)
@@ -268,36 +248,6 @@ class MatchExpression(Expression):
         raise errors.MatchError(arg)
 
 
-class ExecutionExpression(Expression):
-    """Expression that executes all previous once it is found on a pipeline
-
-    Args:
-        args: any arguments to run on the pipeline
-        kwargs: any key-word arguments to run on the pipeline.
-
-    Raises:
-        NotImplementedError: when `>>` is used after it.
-    """
-
-    def __init__(self, *args, **kwargs):
-        op = Operation(lambda *a, **kwd: a[0] if len(a) > 0 else None)
-        super().__init__(f=op)
-        self.__args = args
-        self.__kwargs = kwargs
-
-    def __call__(self, *args, **kwargs):
-        """Computes value of most recent expression, using args on object plus any new ones"""
-        return self._run_prev_expns(*self.__args, *args, **self.__kwargs, **kwargs)
-
-    def __rshift__(self, nxt: Any):
-        """rshift is not supported for this.
-
-        This is a terminal expression that expects no other expression
-        after it on the pipeline.
-        """
-        raise NotImplementedError("terminal pipeline expression: `>>` not supported")
-
-
 class Operation:
     """A computation.
 
@@ -313,7 +263,7 @@ class Operation:
         else:
             self.__f = func
 
-    def __call__(self, *args: Any, **kwargs: "Context") -> Any:
+    def __call__(self, *args: Any, **kwargs: Any) -> Any:
         """Applies the logic attached to this operation and returns output.
 
         Args:
@@ -334,29 +284,11 @@ def _get_func_signature(func: Callable):
         return signature(func.__call__)
 
 
-def to_expn(v: Union["Expression", "Assignment", Callable, Any]) -> "Expression":
+def to_expn(v: Union["Expression", Callable, Any]) -> "Expression":
     """Converts a Callable or Expression into an Expression"""
     if isinstance(v, Expression):
         return v
-    elif isinstance(v, Assignment):
-        # update the context
-        return Expression(
-            Operation(lambda *args, **kwargs: Context(**kwargs, **dict(v)))
-        )
     elif isinstance(v, Callable):
         return Expression(Operation(v))
     # return a noop expression
     return Expression(Operation(func=lambda: v))
-
-
-def _append_expn(
-    first: Union["Expression", "Assignment", Callable, Any],
-    other: Union["Expression", "Assignment", Callable, Any],
-):
-    """Returns a new combined Expression where the current expression runs before the passed expression"""
-    other = to_expn(other)
-    first = to_expn(first)
-
-    other.append_prev_expns(*first._queue, first)
-    first.clear_prev_expns()
-    return other
